
      // This is a 1,3,9,27 ratio tranformer build. Total voltage out is 40 units.
      // we want 240V, so it's 6V/unit at full power
      // Good inverter design has some primary voltage headroom, maybe 20%
      // so let's set normal low voltage cutoff at 48.5V, this will be our 80% power setting
      // 80% of 40 units = 32 units.
      // 48.5V primary = 34.3V AC RMS primary voltage.
      // 240V needed, can only drive it to 32 units. This means 240/32 V per unit = 7.5 V/unit
      // the transformer secondary output voltage will be
      // 7.5 x 1, 7.5 x 3, 7.5 x 9, 7.5 x 27
      // 7.5V AC rms, 22.5V AC, 67.5V AC, 202.5V AC
      // All secondary windings need to be capable of the intended AC output current.
      // Wind the primaries to suit the above output voltages...

#include <SPI.h>

#define ADC_OUT  0          // channel for AC output sample

#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))

volatile long gc;
int sinetab[360];
uint8_t amp2trinary4_pos[41];
uint8_t amp2trinary4_neg[41];
uint8_t trit[3] = {4,0,1};        // not compliment logic, so -1 = B10, 0 = B00, +1 = B01

int vv;                      

uint16_t pcount;

uint8_t uf,oen,dbounce,init_done;
uint16_t vpwr;
bool working;

double Input, Output, Setpoint;
double errSum, lastErr,error,dErr;
double kp, ki, kd;
float ac_output,ac_setpoint,pwr;

int sst;

void setup() 
  {  
  int i,j,k,m,t1,t2,t3,t4;
  float t, u;
  Serial.begin(115200);
    
  TIMSK0 = 0;             // turn off timer 0, it adds jitter to the other interrupts. not needed. it gives millis() etc.
  pinMode(13,OUTPUT);      // LED PORTB 7
  pinMode(9,OUTPUT);      // PORTH 6
  pinMode(8,INPUT);       // continuous on/off pin, driven by nano2
  DDRA = 0xff;            // all 8 bits to be output.
  DDRC = 0xff;           
  DDRK = 0xff;            
  
  TCCR1A =  _BV(WGM11);   // a nice 20kHz timer, which calls the below interrupt
  TCCR1B = _BV(WGM13) | _BV(WGM12) | _BV(CS10);
  ICR1 = 887;
  TIMSK1 |= (1 << TOIE1); // enable the interrupt
  

  for(t=0.0,i=0; i < 180; i++,t +=   1.0 * 3.14159/180.0)
    {   
    u = 40.0 * sin(t);   
    sinetab[i] = (int)(u + 0.5);  
    }      
 
  // set up the 4 inverter bridge control signalling
  // will use this later to determine the signalling for a selected output voltage
  
  amp2trinary4_pos[0] = 0; 
  amp2trinary4_neg[0] = 0; 
  t2=t3=t4=0;
  t1 = 1;
  for(i=1; i <= 40; i++)
    {
    k = 0;  
    m = 0;
    if (t1 ==  1) k += B00000001;
    if (t1 == -1) k += B00000010;
    if (t2 ==  1) k += B00000100;
    if (t2 == -1) k += B00001000;
    if (t3 ==  1) k += B00010000;
    if (t3 == -1) k += B00100000;
    if (t4 ==  1) k += B01000000;
    if (t4 == -1) k += B10000000;

    if (t1 == -1) m += B00000001;
    if (t1 ==  1) m += B00000010;
    if (t2 == -1) m += B00000100;
    if (t2 ==  1) m += B00001000;
    if (t3 == -1) m += B00010000;
    if (t3 ==  1) m += B00100000;
    if (t4 == -1) m += B01000000;
    if (t4 ==  1) m += B10000000;
    amp2trinary4_pos[i] = k; 
    amp2trinary4_neg[i] = m; 
    t1++; 
    if (t1 > 1) 
      {
        t1 = -1;
        t2++;
        if (t2 > 1)
          {
            t2 = -1;
            t3++; 
            if (t3 > 1)
              {
                t3 = -1;
                t4++;
              }
          }      
      }   
    }
  init_vars();
  }


// 20kHz basic clock
ISR(TIMER1_OVF_vect)        
  { 
  int a,p;
  pcount++;if (pcount >= 360) {pcount = 0; uf=1;}

  if (pcount < 180) sbi(PORTH,6); else cbi(PORTH,6);

  p = pcount + vv;
  if (p > 360)    p -= 360;
  if (p <= 180)
    {
      a = (sinetab[p] * vpwr) >> 8;
      PORTK = amp2trinary4_pos[a];
    }
  else
    {
      a = (sinetab[p - 180] * vpwr) >> 8;
      PORTK = amp2trinary4_neg[a];
    }
    
  p = pcount + 120;
  while (p > 360)    p -= 360;
  if (p <= 180)
    {
      a = (sinetab[p] * vpwr) >> 8;
      PORTA = amp2trinary4_pos[a];
    }
  else
    {
      a = (sinetab[p - 180] * vpwr) >> 8;
      PORTA = amp2trinary4_neg[a];
    }

  p = pcount + 240;
  while (p > 360)    p -= 360;
  if (p <= 180)
    {
      a = (sinetab[p] * vpwr) >> 8;
      PORTC = amp2trinary4_pos[a];
    }
  else
    {
      a = (sinetab[p - 180] * vpwr) >> 8;
      PORTC = amp2trinary4_neg[a];
    }
  
  }


void do_pid()
{   
   double timeChange = 0.01;    // always called at 100Hz, so dT will be 10msec..
   error = Setpoint - Input;
   errSum += (error * timeChange);
   dErr = (error - lastErr) / timeChange;
   Output = kp * error + ki * errSum + kd * dErr;
   lastErr = error;
}


void loop() 
  {
  static float ch0;   
  ac_output = (float)analogRead(ADC_OUT) / 1024.0; // scale 0 - 5V to 0.0 to 1.0
  
  //ac_output = 0.2;   // for now

  
  if (uf == 1)  // only execute this once each 50Hz 
    {
      uf=0;  
      ac_setpoint = 0.55;       // for same Vfb as EG8010. change AC output voltage via Vfb voltage divider trimpot
      if (oen == 1) sst++;      // slow start. If output is enabled, keep starting up. sst = output pwm% or approx. voltage.
      if (sst > 251)sst = 251;  // stop increasing when sst has got to the end of slow start. when = 251, this enables PID. See below..
      if (oen == 0) sst--;      // if stopping, slow stop
      if (sst <= 0)             // once fully stopped, no more changes to sst needed
        sst = 0;

      if(sst > 0 && sst < 250)     // slow start timer counts to 250 with gentle ramp up, then switches over to PID control. (gulp!)
        {                          // sst can increase or decrease, giving the slow start and slow stop.
        if (oen == 1)              // slow stop? I think of it as de-gauss
          { 
          if(ac_output < ac_setpoint)
            pwr += 1.0/250.0;     // bang-bang control, should AC output reach setpoint prior to PID.
          else                    // During slow start need to ensure output is tracking setpoint
            pwr -= 1.0/250.0;     // should output get close enough to matter.
          }
        else
          pwr -= 1.0 /250.0;      // this will execute in slow stop
        }
        
      if (sst == 251)       // once sst = 251, run PID. the above bang-bang control will NOT have executed, because sst = 251
        {                   
        Setpoint = ac_setpoint;     
        Input = ac_output;  
        do_pid();
        pwr = pwr + Output;
        }
        
            
      if (pwr < 0.01) pwr = 0.01;       // prevent it going too low for some reason
      if (pwr > 0.99) pwr = 0.99;     // and too high      
      vpwr = int(256.0 * pwr);        // scale and convert to int, ready for fast maths in the 20kHz interrupt  
      //Serial.print(vpwr);      Serial.print( "    ");      Serial.print(oen);      Serial.print("  ");      Serial.println(sst);
      check_switch_cont(); 
    }
  }


void init_vars()      // called prior to setting oen = 1, as well as from setup()
  {
  sst = 0;      // slow start counter 0 - 251
  pcount = 0;   // counter for sine wave lookup table
  uf = 0;       // set to 1 to enable AC output voltage feedback code to execute
  pwr = 0.0;    // PWM duty cycle analog, 0.0 to 1.0
  vpwr = 0;      // integer equivalent of pwr, 0 to PPWM
  dbounce = 0;  // debounce counter for input switch 
  kp = 0.1;     // set up PID for AC output voltage control
  ki = 0.0;
  kd = 0.01;
  errSum = 0.0;        // initialise PID
  lastErr = 0.0; 
  }


void check_switch_cont()
{   // called at 100 Hz
  if ((PINH & 0x20) == 0x20)     
    {
    if (init_done == 0)
      {
      init_vars();
      init_done = 1;
      }  
    oen = 1;      // slow start
    sbi(PORTB,7); // and light LED on D12, pin 15 of nano
    }

  // scr shutdown - without a dummy connection of D6 to ground, inverter wont run.
  //if (((PINB & 0x01) == 0) || ((PIND & 0x40) == 0x40)) 
  // for testing, do not check scr shutdown pin, just nano2 on/off
  if (((PINH & 0x20) == 0) )       
    {
    init_done = 0;
    oen = 0;    // slow stop
    cbi(PORTB,7); // and blank LED on D12, pin 15 of nano
    }
          
}
